@@ -2,7 +2,7 @@
  *
  *  Wireless daemon for Linux
  *
- *  Copyright (C) 2018  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2018-2019  Intel Corporation. All rights reserved.
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Lesser General Public
@@ -20,17 +20,13 @@
  *
  */
 
-struct property_value_options {
-	const char *value_str;
-	const void *value;
-};
-
-static const struct property_value_options properties_on_off_opts[] = {
-	{ "on",  (void *) true },
-	{ "off", (void *) false },
-	{ }
-};
+const char *properties_on_off_opts[3];
+const char *properties_yes_no_opts[3];
 
 bool properties_builder_append_on_off_variant(
 					struct l_dbus_message_builder *builder,
 					const char *value_str);
+
+bool properties_builder_append_yes_no_variant(
+					struct l_dbus_message_builder *builder,
+					const char *value_str);
