@@ -2,7 +2,7 @@
  *
  *  Wireless daemon for Linux
  *
- *  Copyright (C) 2015-2016  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2015-2019  Intel Corporation. All rights reserved.
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Lesser General Public
@@ -1931,19 +1931,38 @@ struct verify_data {
 	verify.expected_len = sizeof(e)
 
 static void verify_handshake_event(struct handshake_state *hs,
-		enum handshake_event event, void *event_data, void *user_data)
+					enum handshake_event event,
+					void *user_data, ...)
 {
 	struct verify_data *data = user_data;
+	va_list args;
+
+	va_start(args, user_data);
 
 	switch (event) {
 	case HANDSHAKE_EVENT_FAILED:
-		assert(l_get_u16(event_data) ==
-				MMPDU_REASON_CODE_IEEE8021X_FAILED);
+		assert(va_arg(args, int) == MMPDU_REASON_CODE_IEEE8021X_FAILED);
 		data->eapol_failed = true;
 		break;
+	case HANDSHAKE_EVENT_EAP_NOTIFY:
+	{
+		const struct wsc_credential *cred;
+
+		assert(va_arg(args, unsigned int) ==
+			EAP_WSC_EVENT_CREDENTIAL_OBTAINED);
+
+		cred = va_arg(args, const struct wsc_credential *);
+		assert(!memcmp(cred, data->expected_creds + data->cur_cred,
+						sizeof(struct wsc_credential)));
+
+		data->cur_cred += 1;
+		break;
+	}
 	default:
 		break;
 	}
+
+	va_end(args);
 }
 
 static int verify_8021x(uint32_t ifindex,
@@ -1963,22 +1982,6 @@ static int verify_8021x(uint32_t ifindex,
 	return 0;
 }
 
-static void verify_credential(unsigned int event, const void *event_data,
-					void *user_data)
-{
-	struct verify_data *data = user_data;
-	const struct wsc_credential *cred;
-
-	if (event != EAP_WSC_EVENT_CREDENTIAL_OBTAINED)
-		assert(false);
-
-	cred = event_data;
-	assert(!memcmp(cred, data->expected_creds + data->cur_cred,
-						sizeof(struct wsc_credential)));
-
-	data->cur_cred += 1;
-}
-
 static void wsc_test_pbc_handshake(const void *data)
 {
 	static uint8_t ap_address[] = { 0x24, 0xa2, 0xe1, 0xec, 0x17, 0x04 };
@@ -2002,11 +2005,8 @@ static void wsc_test_pbc_handshake(const void *data)
 
 	__eapol_set_tx_packet_func(verify_8021x);
 	__eapol_set_tx_user_data(&verify);
-	eapol_sm_set_user_data(sm, &verify);
-	eapol_sm_set_event_func(sm, verify_credential);
 
 	settings = l_settings_new();
-	l_settings_set_uint(settings, "EAP", "mtu", 1400);
 	l_settings_set_string(settings, "Security", "EAP-Identity",
 					"WFA-SimpleConfig-Enrollee-1-0");
 	l_settings_set_string(settings, "Security", "EAP-Method", "WSC");
@@ -2040,7 +2040,6 @@ static void wsc_test_pbc_handshake(const void *data)
 	l_settings_set_string(settings, "WSC", "IV2",
 					"4e3a4cf088176989e148d4c10b96e8fd");
 
-	__eap_set_config(settings);
 	handshake_state_set_8021x_config(hs, settings);
 	eapol_start(sm);
 
@@ -2112,11 +2111,8 @@ static void wsc_test_retransmission_no_fragmentation(const void *data)
 
 	__eapol_set_tx_packet_func(verify_8021x);
 	__eapol_set_tx_user_data(&verify);
-	eapol_sm_set_user_data(sm, &verify);
-	eapol_sm_set_event_func(sm, verify_credential);
 
 	settings = l_settings_new();
-	l_settings_set_uint(settings, "EAP", "mtu", 1400);
 	l_settings_set_string(settings, "Security", "EAP-Identity",
 					"WFA-SimpleConfig-Enrollee-1-0");
 	l_settings_set_string(settings, "Security", "EAP-Method", "WSC");
@@ -2150,7 +2146,6 @@ static void wsc_test_retransmission_no_fragmentation(const void *data)
 	l_settings_set_string(settings, "WSC", "IV2",
 					"4e3a4cf088176989e148d4c10b96e8fd");
 
-	__eap_set_config(settings);
 	handshake_state_set_8021x_config(hs, settings);
 	eapol_start(sm);
 
